// Global marking (PlaceMarking for each place) -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "GlobalMarking.h"
#include "Net.h"
#include "Place.h"
#include "Valuation.h"
#include "Marking.h"
#include "LeafValue.h"

#include <assert.h>

/** @file GlobalMarking.C
 * Assignment of markings for each place
 */

GlobalMarking::GlobalMarking (const class Net& net) :
#ifndef NDEBUG
  myNet (net),
#endif // NDEBUG
  myErroneous (false),
  mySize (net.getNumPlaces ()),
  myPlaceMarkings (mySize ? new class PlaceMarking[mySize] : 0)
{
  for (unsigned i = mySize; i--; )
    myPlaceMarkings[i].setPlace (&net.getPlace (i));
}

GlobalMarking::GlobalMarking (const class GlobalMarking& old) :
#ifndef NDEBUG
  myNet (old.myNet),
#endif // NDEBUG
  myErroneous (old.myErroneous),
  mySize (old.mySize),
  myPlaceMarkings (mySize ? new class PlaceMarking[mySize] : 0)
{
  for (unsigned i = 0; i < mySize; i++)
    myPlaceMarkings[i] = old.myPlaceMarkings[i];
}

class GlobalMarking&
GlobalMarking::operator= (const class GlobalMarking& old)
{
  assert (&myNet == &old.myNet);
  assert (mySize == old.mySize);
  myErroneous = old.myErroneous;
  for (unsigned i = 0; i < mySize; i++)
    myPlaceMarkings[i] = old.myPlaceMarkings[i];
  return *this;
}

GlobalMarking::~GlobalMarking ()
{
  delete[] myPlaceMarkings;
}

class PlaceMarking&
GlobalMarking::operator[] (const class Place& place)
{
  if (place.getNet ().getParent ())
    for (unsigned i = mySize; i--; )
      if (myPlaceMarkings[i].getPlace () == &place)
        return myPlaceMarkings[i];
  assert (!place.getNet ().getParent ());
  assert (place.getIndex () < mySize);
  return myPlaceMarkings[place.getIndex ()];
}

bool
GlobalMarking::encode (class BitPacker& buf,
		       const class GlobalMarking& m0,
		       card_t* errorplace) const
{
  // Clear the encoding buffer
  buf.clear ();
  /** Valuation for checking the consistency of implicit places */
  class Valuation valuation;
  valuation.setGlobalMarking (this);
  // Encode each place marking in the global marking
  for (card_t i = mySize; i--; ) {
    /** Current place marking */
    const class PlaceMarking& place = (*this)[i];
    assert (!!place.getPlace ());
    if (place.getPlace ()->isConstant ()
	? m0[i] == place
	: place.encode (buf, &valuation))
      continue;
    if (errorplace) *errorplace = i;
    return false;
  }

  // the buffer must not be empty
  if (!buf.getNumBytes ())
    buf.append (0, 1);

  return true;
}

void
GlobalMarking::encode (class BitPacker& buf,
		       const class Net& net) const
{
  assert (net.getParent () == &myNet);
  assert (net.getNumPlaces () <= mySize);

  const class Net& parent = *net.getParent ();
  card_t i, j;

  // Clear the encoding buffer
  buf.clear ();

  for (i = net.getNumPlaces (); i--; ) {
    const class Place& place = net.getPlace (i);
    for (j = mySize; j--; )
      if (&parent.getPlace (j) == &place)
	break;
    assert (j < mySize);
    if (!place.isConstant () && !(*this)[j].encode (buf, 0))
      assert (false); // this GlobalMarking should have been valid
  }

  // the buffer must not be empty
  if (!buf.getNumBytes ())
    buf.append (0, 1);
}

unsigned
GlobalMarking::decode (const class GlobalMarking& m0,
		       const word_t* data,
		       unsigned propBits)
{
  card_t i;
  class BitUnpacker buf (data);
  // Decode each place marking in the global marking
  for (i = mySize; i--; )
    (*this)[i].decode (buf);
  computeImplicit (m0);
  return propBits ? buf.extract (propBits) : 0;
}

void
GlobalMarking::decode (const class Net& net,
		       const word_t* data)
{
  assert (net.getParent () == &myNet);
  assert (net.getNumPlaces () <= mySize);
  const class Net& parent = *net.getParent ();
  card_t i, j;
  class BitUnpacker buf (data);
  for (i = net.getNumPlaces (); i--; ) {
    const class Place& place = net.getPlace (i);
    for (j = mySize; j--; )
      if (&parent.getPlace (j) == &place)
	break;
    assert (j < mySize);
    class PlaceMarking& p = (*this)[j];
    p.clear ();
    p.decode (buf);
  }
  computeImplicit (*parent.getInitMarking ());
}

void
GlobalMarking::computeImplicit (const class GlobalMarking& m0)
{
  class Valuation valuation;
  valuation.setGlobalMarking (this);
  for (card_t i = 0; i < mySize; i++) {
    class PlaceMarking& p = (*this)[i];
    const class Place& place = *p.getPlace ();
    if (place.isConstant ()) {
      assert (p.empty ());
      p = m0[i];
    }
    else if (place.isImplicit ()) {
      const class Marking* expr = place.getInitMarking ();
      assert (!!expr);
      p.clear ();
      if (!expr->add (p, 1, valuation))
	assert (false);
    }
  }
}

bool
GlobalMarking::eval (const class Expression& cond) const
{
  class Valuation valuation;
  valuation.setGlobalMarking (this);
  bool holds = false;
  if (class Value* value = cond.eval (valuation)) {
    assert (value->getKind () == Value::vLeaf);
    assert (valuation.isOK ());
    holds = bool (static_cast<class LeafValue&>(*value));
    delete value;
  }
  return holds;
}

#include "Printer.h"
void
GlobalMarking::display (const class Printer& printer) const
{
  printer.printRaw (isErroneous () ? "erroneous state " : "state ");
  printer.delimiter ('(')++;

  for (unsigned i = 0; i < mySize; i++) {
    if (!myPlaceMarkings[i].empty () &&
	!myPlaceMarkings[i].getPlace ()->isSuppressed ()) {
      printer.linebreak ();
      myPlaceMarkings[i].display (printer);
    }
  }

  printer--.linebreak ();
  printer.delimiter (')');
}
